/* Front-end language hooks for the brainfuck language.
   Copyright (c) 2013 Free Software Foundation, Inc.
   Copyright (c) 2014 Christopher Smith

This file is part of gbf.

gbf is free software: you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

gbf is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
gbf. If not, see <http://www.gnu.org/licenses/>. */

#include "config.h"
#include "system.h"
#include "ansidecl.h"
#include "coretypes.h"
#include "tree.h"
#include "tree-iterator.h"
#include "dumpfile.h"
#include "gimple.h"
#include "cgraph.h"
#include "opts.h"
#include "options.h"
#include "flags.h"
#include "diagnostic-core.h"
#include "toplev.h"
#include "langhooks.h"
#include "langhooks-def.h"
#include "debug.h"
#include "vec.h"

/* Required structures. */

struct GTY(()) lang_decl
{
  char dummy;
};

struct GTY(()) lang_type
{
  char dummy;
};

struct GTY(()) language_function
{
  char dummy;
};

struct GTY(()) lang_identifier
{
  struct tree_identifier common;
};

union GTY((desc ("0"),
           chain_next ("(union lang_tree_node *) TREE_CHAIN (&%h.generic)")))
      lang_tree_node
{
  union tree_node GTY((tag ("0"), desc ("tree_node_structure (&%h)"))) generic;
};

/* Internal structures. */

struct GTY(()) bf_op
{
  char code;
  source_location loc;
};

/* Global variables. */

tree global_decls = NULL_TREE;

/* Language hook functions. */

unsigned int
bf_option_lang_mask (void)
{
  return CL_brainfuck;
}

bool
bf_init (void)
{
  /* Create default types like int. */
  build_common_tree_nodes (false, false);
  return true;
}

static bool
read_file_check_syntax (const char *fname, vec<struct bf_op> *file_data)
{
  FILE *in = fopen (fname, "r");
  if (in == NULL)
    fatal_error ("Can't open input file %s", fname);

  linemap_add (line_table, LC_ENTER, 0, fname, 1);
  unsigned int line = 0, column = 0;
  linemap_line_start (line_table, ++line, 0);
  struct bf_op cur_instr;
  bool result = true;

  vec<source_location> open_brackets = vNULL;
  int c;
  while ((c = fgetc (in)) != EOF)
    {
      column++;
      switch (c)
        {
        case '\n':
          linemap_line_start (line_table, ++line, 0);
          column = 0;
          break;

        case '>': case '<':
        case '+': case '-':
        case '[': case ']':
        case ',': case '.':
          cur_instr.code = c;
          cur_instr.loc = linemap_position_for_column (line_table, column);

          if (c == '[')
            open_brackets.safe_push (cur_instr.loc);
          else if (c == ']')
            {
              if (open_brackets.length () == 0)
                {
                  error_at (cur_instr.loc, "Unmatched ]");
                  result = false;
                }
              else
                open_brackets.pop ();
            }

          file_data->safe_push (cur_instr);
          break;
        }
    }

  if (open_brackets.length () != 0)
    {
      unsigned int ix;
      source_location loc;
      FOR_EACH_VEC_ELT (open_brackets, ix, loc)
        error_at (loc, "Unmatched [");
      result = false;
    }

  fclose (in);
  return result;
}

static tree
build_alloc_expr (tree p_decl)
{
  /* Build the function declaration. */
  tree fn_type = build_function_type (ptr_type_node, NULL_TREE);
  tree fn_decl = build_decl (BUILTINS_LOCATION, FUNCTION_DECL,
                             get_identifier ("calloc"), fn_type);

  DECL_EXTERNAL (fn_decl) = true;
  TREE_PUBLIC (fn_decl) = true;

  /* Build the call expression. */
  tree call_expr = build_call_expr (fn_decl, 2,
                                    build_int_cst (size_type_node, 30000),
                                    build_int_cst (size_type_node, 1));

  /* Assign the result to the variable. */
  tree assign_expr = build2 (MODIFY_EXPR, ptr_type_node, p_decl, call_expr);

  return assign_expr;
}

static void
append_brainfuck_code (vec<struct bf_op> *code, tree p_decl, tree *main_stmts)
{
  vec<tree> blocks = vNULL;
  blocks.safe_push (*main_stmts);

  tree one = build_int_cst (char_type_node, 1);

  tree deref_expr = build1 (INDIRECT_REF, char_type_node, p_decl);

  tree test_expr = build2 (EQ_EXPR, integer_type_node, deref_expr,
                           build_int_cst (char_type_node, 0));

  tree getchar_type = build_function_type (integer_type_node, NULL_TREE);
  tree getchar_decl = build_decl (BUILTINS_LOCATION, FUNCTION_DECL,
                                  get_identifier ("getchar"), getchar_type);
  tree call_get = build_call_expr (getchar_decl, 0);

  tree putchar_type = build_function_type (integer_type_node, NULL_TREE);
  tree putchar_decl = build_decl (BUILTINS_LOCATION, FUNCTION_DECL,
                                  get_identifier ("putchar"), putchar_type);

  int ix;
  struct bf_op *cmd;
  FOR_EACH_VEC_ELT (*code, ix, cmd)
    {
      tree stmt;
      switch (cmd->code)
        {
        case '>':
          stmt = build2 (PREINCREMENT_EXPR, ptr_type_node, p_decl, one);
          break;
        case '<':
          stmt = build2 (PREDECREMENT_EXPR, ptr_type_node, p_decl, one);
          break;
        case '+':
          stmt = build2 (PREINCREMENT_EXPR, char_type_node, deref_expr, one);
          break;
        case '-':
          stmt = build2 (PREDECREMENT_EXPR, char_type_node, deref_expr, one);
          break;
        case '[':
          blocks.safe_push (alloc_stmt_list ());
          stmt = build1 (EXIT_EXPR, void_type_node, test_expr);
          break;
        case ']':
          stmt = build1 (LOOP_EXPR, void_type_node, blocks.pop ());
          break;
        case ',':
          stmt = build2 (MODIFY_EXPR, char_type_node, deref_expr, call_get);
          break;
        case '.':
          stmt = build_call_expr (putchar_decl, 1, deref_expr);
          break;
        default:
          gcc_unreachable ();
        }
      append_to_statement_list (stmt, &blocks.last ());
    }

  *main_stmts = blocks.last ();
}

static tree
build_main_function (vec<struct bf_op> *code)
{
  /* Create the main () function. */
  tree main_fn_type = build_function_type (integer_type_node, NULL_TREE);
  tree main_fn_decl = build_decl (BUILTINS_LOCATION, FUNCTION_DECL,
                                  get_identifier ("main"), main_fn_type);

  DECL_CONTEXT (main_fn_decl) = NULL_TREE;
  TREE_STATIC (main_fn_decl) = true;
  TREE_PUBLIC (main_fn_decl) = true;
  DECL_ARGUMENTS (main_fn_decl) = NULL_TREE;

  /* Return value for main (). */
  tree main_ret = build_decl (BUILTINS_LOCATION, RESULT_DECL, NULL_TREE,
                              TREE_TYPE (main_fn_type));
  DECL_CONTEXT (main_ret) = main_fn_decl;
  DECL_ARTIFICIAL (main_ret) = true;
  DECL_IGNORED_P (main_ret) = true;

  DECL_RESULT (main_fn_decl) = main_ret;

  /* Variable declaration - just the data pointer. */
  tree p_decl = build_decl (BUILTINS_LOCATION, VAR_DECL, NULL_TREE,
                            build_pointer_type (char_type_node));

  tree bl = build_block (p_decl, NULL_TREE, main_fn_decl, NULL_TREE);
  DECL_INITIAL (main_fn_decl) = bl;
  TREE_USED (bl) = true;

  /* The bind expression contains the statements to execute. */
  tree bind = build3 (BIND_EXPR, void_type_node, BLOCK_VARS (bl), NULL_TREE,
                      bl);
  TREE_SIDE_EFFECTS (bind) = true;

  /* List of statements in the main () function. */
  tree main_stmts = alloc_stmt_list ();

  append_to_statement_list (build_alloc_expr (p_decl), &main_stmts);

  append_brainfuck_code (code, p_decl, &main_stmts);

  /* return 0; */
  tree main_set_ret = build2 (MODIFY_EXPR, TREE_TYPE (main_ret), main_ret,
                              build_int_cst (integer_type_node, 0));
  TREE_USED (main_set_ret) = true;
  tree main_ret_expr = build1 (RETURN_EXPR, void_type_node, main_set_ret);
  append_to_statement_list (main_ret_expr, &main_stmts);

  BIND_EXPR_BODY (bind) = main_stmts;
  DECL_SAVED_TREE (main_fn_decl) = bind;

  return main_fn_decl;
}

/* Language hook to read and parse the source file. */
void
bf_parse_file (void)
{
  /* Read in the source file. */
  vec<struct bf_op> file_data = vNULL;
  if (!read_file_check_syntax (in_fnames[0], &file_data))
    return;

  global_decls = chainon (build_main_function (&file_data), global_decls);

  /* Dump out the translation unit. */
  FILE *tu_stream = dump_begin (TDI_tu, NULL);
  if (tu_stream)
    {
      dump_node (global_decls, 0, tu_stream);
      dump_end (TDI_tu, tu_stream);
    }

  /* Prepare declarations for middle-end. */
  for (tree t = global_decls; t; t = TREE_CHAIN (t))
    {
      gimplify_function_tree (t);
      cgraph_finalize_function (t, false);
    }
}

/* Unused; terminate with ICE. */
bool
bf_global_bindings_p (void)
{
  gcc_unreachable ();
}

/* Unused; terminate with ICE. */
tree
bf_pushdecl (tree decl ATTRIBUTE_UNUSED)
{
  gcc_unreachable ();
}

/* Return the global declarations. */
tree
bf_getdecls (void)
{
  return global_decls;
}

tree
bf_type_for_mode (enum machine_mode mode, int unsignedp)
{
  if (mode == QImode)
    return unsignedp ? unsigned_char_type_node : signed_char_type_node;

  gcc_unreachable ();
}

tree
bf_type_for_size (unsigned precision, int unsignedp)
{
  if (precision == 8)
    return unsignedp ? unsigned_char_type_node : signed_char_type_node;

  gcc_unreachable ();
}

/* Language hooks time! */

#undef LANG_HOOKS_NAME
#undef LANG_HOOKS_OPTION_LANG_MASK
#undef LANG_HOOKS_INIT
#undef LANG_HOOKS_PARSE_FILE
#undef LANG_HOOKS_GLOBAL_BINDINGS_P
#undef LANG_HOOKS_PUSHDECL
#undef LANG_HOOKS_GETDECLS
#undef LANG_HOOKS_TYPE_FOR_MODE
#undef LANG_HOOKS_TYPE_FOR_SIZE

#define LANG_HOOKS_NAME              "GNU brainfuck"
#define LANG_HOOKS_OPTION_LANG_MASK  bf_option_lang_mask
#define LANG_HOOKS_INIT              bf_init
#define LANG_HOOKS_PARSE_FILE        bf_parse_file
#define LANG_HOOKS_GLOBAL_BINDINGS_P bf_global_bindings_p
#define LANG_HOOKS_PUSHDECL          bf_pushdecl
#define LANG_HOOKS_GETDECLS          bf_getdecls
#define LANG_HOOKS_TYPE_FOR_MODE     bf_type_for_mode
#define LANG_HOOKS_TYPE_FOR_SIZE     bf_type_for_size

struct lang_hooks lang_hooks = LANG_HOOKS_INITIALIZER;

/* GGC stuff. */

#include "gt-bf-bf1.h"
#include "gtype-bf.h"
